# -*- coding: utf-8 -*-
"""
This script determines the MOS design parameters (ID, W, and L)
for a CS stage driven from am capacitive source, on grounds of
design limits for the floor noise and for the 1/f noise.
"""

from SLiCAP import *
from extras import *

filename = 'noiseCSN'

# Specifications: alternatively you may use specification items (specItem() objects), store them in a csv file
# and use them throughout your design

Se_floor_max = 100e-18             # max antenna-referred voltage floor noise V2/m2/Hz
fc_max       = 200e3               # corner frequency of low-frequency noise at maximum floor noise
C_A          = 6e-12               # antenna capacitance
L_A          = 0.5                 # antenna length
G            = 1                   # gain of the active antenna V/(V/m)
C_F          = C_A/G               # feedback capacitance

L            = 0.22e-6              # Selected channel length
VGS_max      = 1.8                 # Maximum VGS budget for input stage (absolute value)

i1 = instruction()
i1.setCircuit(filename+'.cir')
i1.setSimType('symbolic')
i1.setGainType('vi')
i1.setSource('V1')
i1.setDetector('V_Out')
i1.setDataType('noise')

# define circuit parameters
i1.defPar("L", L)
i1.defPar("C_A", C_A)
i1.defPar("C_F", C_F)
i1.defPar("L_A", L_A)

# store the symbolic definitions of f_T, IC, c_iss and g_m
fT   = i1.getParValue('f_T_XU1')
IC   = i1.getParValue("IC_XU1")
ciss = i1.getParValue("c_iss_XU1")
gm   = i1.getParValue("g_m_XU1")

# obtain a symbolic expression for the antenna-referred input noise
noise_result = i1.execute()
inoise = noise_result.inoise

# Express the input noise in terms of 'c_iss'
# To this end we write g_m as a function of c_iss and f_T
g_m = sp.solve(fT-sp.Symbol('f_T_XU1'), sp.Symbol('g_m_XU1'))[0]

# Now substitute this function into the expression for inoise
inoise = inoise.subs(sp.Symbol('g_m_XU1'), g_m)

# The gate current is zero:
inoise = inoise.subs(sp.Symbol('IG'), 0)

# For determination of the floor noise: define AF as positive
# For NMOS devices
inoise = sp.simplify(inoise.subs(sp.Symbol('AF_N18'), sp.Symbol('AF_N18', positive=True)))
# For PMOS devices
inoise = sp.simplify(inoise.subs(sp.Symbol('AF_P18'), sp.Symbol('AF_P18', positive=True)))

# The floor noise is found as the noise for f --> infinity
floornoise = sp.factor(sp.limit(inoise,ini.frequency,'oo'))

# Reset assumption for AF_N18
floornoise    = clearAssumptions(floornoise)
inoise        = clearAssumptions(inoise)
lowfreqnoise  = sp.simplify(inoise-floornoise)

# Design the transistor
IC_init  = 1
IC_found = .5
error = False
i = 0
while sp.Abs(1-IC_found/IC_init) > 0.01 and not error:
    print("IC", sp.N(IC_found,3))
    IC_init = IC_found
    # A minimum value for c_iss follows from the 1/f noise
    # At the corner frequency, the low-frequency noise can be as large as the
    # specified floor noise
    ciss_min_sym = sp.solve(lowfreqnoise.subs(ini.frequency, fc_max) - Se_floor_max, sp.Symbol('c_iss_XU1'))[0]
    i1.defPar("IC_XU1", IC_init)
    ciss_min_num = sp.N(fullSubs(ciss_min_sym, i1.parDefs))

    # The floor noise can be expressed in terms of g_m and c_iss and the design limit for the floor noise
    gm_ciss = sp.solve(floornoise.subs(sp.Symbol("f_T_XU1"), fT) - Se_floor_max, sp.Symbol("g_m_XU1"))[0]
    i1.defPar("c_iss_XU1", ciss_min_num)
    gm_min = fullSubs(gm_ciss, i1.parDefs)
    i1.defPar("g_m_XU1", gm_min)
    fT_min = sp.N(fullSubs(fT, i1.parDefs))

    # Restore definition of c_iss
    i1.defPar("c_iss_XU1", ciss)

    # Restore definition of g_m
    i1.defPar("g_m_XU1", gm)
    i1.setSimType("numeric")

    # Determine W
    W = sp.solve(i1.getParValue("c_iss_XU1") - ciss_min_num, sp.Symbol("W"))[0]
    i1.defPar("W", W)

    # Determine ID
    ID = sp.solve(i1.getParValue("g_m_XU1") - gm_min)[0]
    i1.defPar("ID", ID)

    # Restore definition of IC
    i1.defPar("IC_XU1", IC)

    # Calculate new value of inversion coefficient
    IC_found = fullSubs(i1.getParValue("IC_XU1"), i1.parDefs)

    # Calculate VGS
    VGS = fullSubs(i1.getParValue("V_GS_XU1"), i1.parDefs)
    i += 1

    # Check VGS limit
    if sp.Abs(VGS) > VGS_max:
        print("VGS requirement cannot be met.")
        error = True
    # Initialize for next loop
    i1.defPar('f_T_XU1', fT)
    i1.defPar("IC_XU1", IC)
    i1.defPar("c_iss_XU1", ciss)
    i1.defPar("g_m_XU1", gm)
    i1.delPar("W")
    i1.delPar("ID")

if not error:
    print("\nFound results ater %s iterations for %s:"%(str(i), filename))
    print("c_iss   :", sp.N(ciss_min_num, 3))
    print("g_m     :", sp.N(gm_min, 3))
    print("L       :", sp.N(L, 3))
    print("W       :", sp.N(W, 3))
    print("ID      :", sp.N(ID, 3))
    print("f_T     :", sp.N(fT_min, 3))
    print("IC      :", sp.N(IC_found, 3))
    print("|VGS|   :", sp.N(sp.Abs(VGS), 3))
    print("|gm/ID| :", sp.N(sp.Abs(gm_min/ID), 3))

# plot the noise spectrum
i1.setSimType("numeric")
i1.defPar("W", W)
i1.defPar("L", L)
i1.defPar("ID", ID)
i1.defPar("IG", 0)
numNoise = i1.execute()
SE = plotSweep("SE", "E-field noise", numNoise, 0.001, 1000, 200, sweepScale="M", funcType="inoise", show=True)

###############################################################################################################
# Compare noise EKV-BSIM for given device and operating point
###############################################################################################################

# MS Windows: define the location of the ngspice.exe"
# ini.ngspiceCMD = '<path to ngspice.exe>'
# Example NMOS operating point info

# Define library and device
refDes   = 'M1'                     # Reference designator of MOS device
if filename == 'noiseCSP':
    DEV  = 'pch'                    # MOS device name
else:
    DEV  = 'nch'
LIB      = '.lib lib/log018.l TT'   # use this CMOS BSIM library with EKV CMOS18.lib
Wf       = 50e-6                    # Finger width
L        = 0.5e-6                   # Channel length
M        = 1+int(W/Wf)              # Number of devices in parallel

####################################################################################################
# Create the device
device = MOS(refDes, LIB, DEV, Wf, L, M)

# Define frequency range for noise analysis
fmin     = 1e3                      # Start frequency
fmax     = 1e9                      # Stop frequency
numDec   = 20                       # Number of points per decade

# Define the N_Channel operating point
VD       = 0.9                      # Drain voltage w.r.t. GND
VP       = 1.8                      # Supply voltage w.r.t. GND

#calculate terminal voltages

if DEV == "pch":
    VS = VP
    VB = VP
    VD = VP - VD
    ID = -ID

else:
    VS = 0
    VB = 0

# plot the voltage noise
plotSvinoise(device, ID, VD, VS, VB, fmin, fmax, numDec, EKVlib = 'CMOS18.lib')